@@ -18,10 +18,10 @@ module Language.C.Analysis.DeclAnalysis (
   mergeOldStyle,
   -- * Dissecting type specs
   canonicalTypeSpec, NumBaseType(..),SignSpec(..),SizeMod(..),NumTypeSpec(..),TypeSpecAnalysis(..),
-  canonicalStorageSpec, StorageSpec(..),hasThreadLocalSpec,
+  canonicalStorageSpec, StorageSpec(..),hasThreadLocalSpec, isTypeDef,
   -- * Helpers
   VarDeclInfo(..),
-  tAttr,mkVarName,getOnlyDeclr,nameOfDecl,
+  tAttr,mkVarName,getOnlyDeclr,nameOfDecl,analyseVarDecl,analyseVarDecl'
 )
 where
 import Language.C.Data.Error
@@ -29,7 +29,9 @@ import Language.C.Data.Node
 import Language.C.Data.Ident
 import Language.C.Pretty
 import Language.C.Syntax
-import Language.C.Analysis.DefTable (TagFwdDecl(..))
+import {-# SOURCE #-} Language.C.Analysis.AstAnalysis (tExpr, ExprSide(..))
+import Language.C.Analysis.DefTable (TagFwdDecl(..), insertType, lookupType)
+import Language.C.Analysis.Export
 import Language.C.Analysis.SemError
 import Language.C.Analysis.SemRep
 import Language.C.Analysis.TravMonad
@@ -51,12 +53,12 @@ tParamDecl :: (MonadTrav m) => CDecl -> m ParamDecl
 tParamDecl (CDecl declspecs declrs node) =
   do declr <- getParamDeclr
      -- analyse the variable declaration
-     (VarDeclInfo name is_inline storage_spec attrs ty declr_node) <- analyseVarDecl True declspecs declr [] Nothing
+     (VarDeclInfo name is_inline storage_spec attrs ty declr_node) <- analyseVarDecl' True declspecs declr [] Nothing
      when (is_inline) $ throwTravError (badSpecifierError node "parameter declaration with inline specifier")
      -- compute storage of parameter (NoStorage, but might have a register specifier)
      storage <- throwOnLeft $ computeParamStorage node storage_spec
      let paramDecl = mkParamDecl name storage attrs ty declr_node
-     handleParamDecl paramDecl
+     -- XXX: we shouldn't modify the deftable here, just analyse and build representation
      return $ paramDecl
   where
   getParamDeclr =
@@ -69,6 +71,7 @@ tParamDecl (CDecl declspecs declrs node) =
     case name of
       NoName -> AbstractParamDecl vd declr_node
       _ -> ParamDecl vd declr_node
+
 -- | a parameter declaration has no linkage and either auto or register storage
 computeParamStorage :: NodeInfo -> StorageSpec -> Either BadSpecifierError Storage
 computeParamStorage _ NoStorageSpec = Right (Auto False)
@@ -77,10 +80,26 @@ computeParamStorage node spec       = Left . badSpecifierError node $ "Bad stora
 
 -- | analyse and translate a member declaration
 tMemberDecls :: (MonadTrav m) => CDecl -> m [MemberDecl]
+-- Anonymous struct or union members
+tMemberDecls (CDecl declspecs [] node) =
+  do let (storage_specs, _attrs, typequals, typespecs, is_inline) =
+           partitionDeclSpecs declspecs
+     when is_inline $ astError node "member declaration with inline specifier"
+     canonTySpecs <- canonicalTypeSpec typespecs
+     ty <- tType True node typequals canonTySpecs [] []
+     case ty of
+       DirectType (TyComp _) _ _ ->
+         return $ [MemberDecl
+                   -- XXX: are these DeclAttrs correct?
+                   (VarDecl NoName (DeclAttrs False NoStorage []) ty)
+                   Nothing node]
+       _ -> astError node "anonymous member has a non-composite type"
+-- Named members
 tMemberDecls (CDecl declspecs declrs node) = mapM (uncurry tMemberDecl) (zip (True:repeat False) declrs)
     where
     tMemberDecl handle_sue_def (Just member_declr,Nothing,bit_field_size_opt) =
-        do var_decl <- analyseVarDecl handle_sue_def declspecs member_declr [] Nothing
+        -- TODO: use analyseVarDecl here, not analyseVarDecl'
+        do var_decl <- analyseVarDecl' handle_sue_def declspecs member_declr [] Nothing
            let (VarDeclInfo name is_inline storage_spec attrs ty declr_node) = var_decl
            --
            checkValidMemberSpec is_inline storage_spec
@@ -88,7 +107,8 @@ tMemberDecls (CDecl declspecs declrs node) = mapM (uncurry tMemberDecl) (zip (Tr
     tMemberDecl handle_sue_def (Nothing,Nothing,Just bit_field_size) =
         do let (storage_specs, _attrs, typequals, typespecs, is_inline) = partitionDeclSpecs declspecs
            storage_spec  <- canonicalStorageSpec storage_specs
-           typ           <- tType handle_sue_def node typequals typespecs [] []
+           canonTySpecs  <- canonicalTypeSpec typespecs
+           typ           <- tType handle_sue_def node typequals canonTySpecs [] []
            --
            return $ AnonBitField typ bit_field_size node
     tMemberDecl _ _ = astError node "Bad member declaration"
@@ -108,16 +128,28 @@ hasThreadLocalSpec _  = False
 
 data VarDeclInfo = VarDeclInfo VarName Bool {- is-inline? -} StorageSpec Attributes Type NodeInfo
 
+analyseVarDecl' :: (MonadTrav m) =>
+                  Bool -> [CDeclSpec] ->
+                  CDeclr -> [CDecl] -> (Maybe CInit) -> m VarDeclInfo
+analyseVarDecl' handle_sue_def declspecs declr oldstyle init_opt =
+  do let (storage_specs, attrs, type_quals, type_specs, inline) =
+           partitionDeclSpecs declspecs
+     canonTySpecs <- canonicalTypeSpec type_specs
+     analyseVarDecl handle_sue_def storage_specs attrs type_quals canonTySpecs inline
+                    declr oldstyle init_opt
+
 -- | analyse declarators
-analyseVarDecl :: (MonadTrav m) => Bool -> [CDeclSpec] -> CDeclr -> [CDecl] -> (Maybe CInit) -> m VarDeclInfo
-analyseVarDecl handle_sue_def declspecs
+analyseVarDecl :: (MonadTrav m) =>
+                  Bool -> [CStorageSpec] -> [CAttr] -> [CTypeQual] ->
+                  TypeSpecAnalysis -> Bool ->
+                  CDeclr -> [CDecl] -> (Maybe CInit) -> m VarDeclInfo
+analyseVarDecl handle_sue_def storage_specs decl_attrs typequals canonTySpecs inline
                (CDeclr name_opt derived_declrs asmname_opt declr_attrs node)
                oldstyle_params init_opt
-    = do let (storage_specs, decl_attrs, typequals, typespecs, inline) = partitionDeclSpecs declspecs
-         -- analyse the storage specifiers
+    = do -- analyse the storage specifiers
          storage_spec  <- canonicalStorageSpec storage_specs
          -- translate the type into semantic representation
-         typ          <- tType handle_sue_def node typequals typespecs derived_declrs oldstyle_params
+         typ          <- tType handle_sue_def node typequals canonTySpecs derived_declrs oldstyle_params
          -- translate attributes
          attrs'       <- mapM tAttr (decl_attrs ++ declr_attrs)
          -- make name
@@ -132,7 +164,6 @@ analyseVarDecl handle_sue_def declspecs
 isTypeDef :: [CDeclSpec] -> Bool
 isTypeDef declspecs = not $ null [ n | (CStorageSpec (CTypedef n)) <- declspecs ]
 
-
 -- * translation
 
 -- | get the type of a /type declaration/
@@ -156,25 +187,34 @@ analyseTypeDecl (CDecl declspecs declrs node)
     where
     analyseTyDeclr (CDeclr Nothing derived_declrs Nothing attrs _declrnode)
         | (not (null storagespec) || inline) = astError node "storage specifier for type declaration"
-        | otherwise                          = tType True node (map CAttrQual (attrs++attrs_decl) ++ typequals)
-                                                     typespecs derived_declrs []
+        | otherwise                          =
+          do canonTySpecs <- canonicalTypeSpec typespecs
+             t <- tType True node (map CAttrQual (attrs++attrs_decl) ++ typequals)
+                   canonTySpecs derived_declrs []
+             case nameOfNode node of
+               Just n -> withDefTable (\dt -> (t, insertType dt n t))
+               Nothing -> return t
         where
         (storagespec, attrs_decl, typequals, typespecs, inline) = partitionDeclSpecs declspecs
     analyseTyDeclr _ = astError node "Non-abstract declarator in type declaration"
 
 
 -- | translate a type
-tType :: (MonadTrav m) => Bool -> NodeInfo -> [CTypeQual] -> [CTypeSpec] -> [CDerivedDeclr] -> [CDecl] -> m Type
-tType handle_sue_def top_node typequals typespecs derived_declrs oldstyle_params
+tType :: (MonadTrav m) => Bool -> NodeInfo -> [CTypeQual] -> TypeSpecAnalysis -> [CDerivedDeclr] -> [CDecl] -> m Type
+tType handle_sue_def top_node typequals canonTySpecs derived_declrs oldstyle_params
     = mergeOldStyle top_node oldstyle_params derived_declrs >>= buildType
     where
-    buildType [] = tDirectType handle_sue_def top_node typequals typespecs
+    buildType [] =
+        tDirectType handle_sue_def top_node typequals canonTySpecs
     buildType (CPtrDeclr ptrquals node : dds) =
         buildType dds >>= buildPointerType ptrquals node
     buildType (CArrDeclr arrquals size node : dds)
         = buildType dds >>= buildArrayType arrquals size node
-    buildType (CFunDeclr ~(Right (params, isVariadic)) attrs node : dds)
-        = buildType dds >>= (liftM FunctionType . buildFunctionType params isVariadic attrs node)
+    buildType (CFunDeclr (Right (params, isVariadic)) attrs node : dds)
+        = buildType dds >>= (liftM  (uncurry FunctionType) . buildFunctionType params isVariadic attrs node)
+    buildType (CFunDeclr (Left _) _ _ : _)
+        -- /FIXME/: this is really an internal error, not an AST error.
+        = astError top_node "old-style parameters remaining after mergeOldStyle"
     buildPointerType ptrquals _node inner_ty
         = liftM (\(quals,attrs) -> PtrType inner_ty quals attrs) (tTypeQuals ptrquals)
     buildArrayType arr_quals size _node inner_ty
@@ -187,19 +227,20 @@ tType handle_sue_def top_node typequals typespecs derived_declrs oldstyle_params
         = do enterPrototypeScope
              params' <- mapM tParamDecl params
              leavePrototypeScope
-             attrs'   <- mapM tAttr attrs
-             return $ case (map declType params',is_variadic) of
-               ([],False) -> FunTypeIncomplete return_ty attrs'  -- may be improved later on
-               ([DirectType TyVoid _],False) -> FunType return_ty [] False attrs'
-               _ -> FunType return_ty params' is_variadic attrs'
+             attrs'  <- mapM tAttr attrs
+             return $ (\t -> (t,attrs')) $
+                case (map declType params',is_variadic) of
+                    ([],False) -> FunTypeIncomplete return_ty  -- may be improved later on
+                    ([DirectType TyVoid _ _],False) -> FunType return_ty [] False
+                    _ -> FunType return_ty params' is_variadic
 
 -- | translate a type without (syntactic) indirections
 -- Due to the GNU @typeof@ extension and typeDefs, this can be an arbitrary type
-tDirectType :: (MonadTrav m) => Bool -> NodeInfo -> [CTypeQual] -> [CTypeSpec] -> m Type
-tDirectType handle_sue_def node ty_quals ty_specs = do
+tDirectType :: (MonadTrav m) =>
+               Bool -> NodeInfo -> [CTypeQual] -> TypeSpecAnalysis -> m Type
+tDirectType handle_sue_def node ty_quals canonTySpec = do
     (quals,attrs) <- tTypeQuals ty_quals
-    canonTySpec <- canonicalTypeSpec ty_specs
-    let baseType ty_name = DirectType ty_name quals
+    let baseType ty_name = DirectType ty_name quals attrs
     case canonTySpec of
         TSNone -> return$ baseType (TyIntegral TyInt)
         TSVoid -> return$ baseType TyVoid
@@ -211,12 +252,10 @@ tDirectType handle_sue_def node ty_quals ty_specs = do
                     Left (floatType,iscomplex) | iscomplex -> TyComplex floatType
                                                | otherwise -> TyFloating floatType
                     Right intType  -> TyIntegral intType
+        TSTypeDef tdr -> return$ TypeDefType tdr quals attrs
         TSNonBasic (CSUType su _tnode)      -> liftM (baseType . TyComp) $ tCompTypeDecl handle_sue_def su
         TSNonBasic (CEnumType enum _tnode)   -> liftM (baseType . TyEnum) $ tEnumTypeDecl handle_sue_def enum
-        TSNonBasic (CTypeDef name t_node)    -> liftM TypeDefType $ typeDefRef t_node name
-        -- TODO: analyse type of expression
-        TSNonBasic (CTypeOfExpr expr _tnode) -> return $ TypeOfExpr expr
-        TSNonBasic (CTypeOfType decl t_node) ->  analyseTypeDecl decl >>= mergeTypeAttributes t_node quals attrs
+        TSType t                             ->  mergeTypeAttributes node quals attrs t
         TSNonBasic _ -> astError node "Unexpected typespec"
 
 -- | Merge type attributes
@@ -225,21 +264,24 @@ tDirectType handle_sue_def node ty_quals ty_specs = do
 --
 -- > /* tyqual attr typeof(type) */
 -- > const typeof(char volatile) x;
-mergeTypeAttributes :: (MonadTrav m) => NodeInfo -> TypeQuals -> [Attr] -> Type -> m Type
+mergeTypeAttributes :: (MonadCError m) => NodeInfo -> TypeQuals -> [Attr] -> Type -> m Type
 mergeTypeAttributes node_info quals attrs typ =
     case typ of
-        DirectType ty_name quals' -> merge quals' [] $ mkDirect ty_name
+        DirectType ty_name quals' attrs' -> merge quals' attrs' $ mkDirect ty_name
         PtrType ty quals' attrs'  -> merge quals' attrs' $ PtrType ty
         ArrayType ty array_sz quals' attrs' -> merge quals' attrs' $ ArrayType ty array_sz
-        FunctionType (FunType return_ty params inline attrs')
-            | not (null attrs) -> astError node_info "type qualifiers for function type"
-            | otherwise        -> return$ FunctionType (FunType return_ty params inline (attrs' ++ attrs))
+        FunctionType (FunType return_ty params inline) attrs'
+-- /FIXME/: This needs review, but checking (null attrs) seems strange here
+--            | not (null attrs) -> astError node_info "type qualifiers for function type"
+--           | otherwise        -> return$ FunctionType (FunType return_ty params inline (attrs' ++ attrs))
+            -> return$ FunctionType (FunType return_ty params inline)  (attrs' ++ attrs)
+        TypeDefType tdr quals' attrs'
+            -> merge quals' attrs' $ TypeDefType tdr
     where
-    mkDirect ty_name quals' attrs' | null attrs' = DirectType ty_name quals'
-                                   | otherwise   = error "_attribute__ s for DirectType"
+    mkDirect ty_name quals' attrs' = DirectType ty_name quals' attrs'
     merge quals' attrs' tyf = return $ tyf (mergeTypeQuals quals quals') (attrs' ++ attrs)
 
-typeDefRef :: (MonadTrav m) => NodeInfo -> Ident -> m TypeDefRef
+typeDefRef :: (MonadCError m, MonadSymtab m) => NodeInfo -> Ident -> m TypeDefRef
 typeDefRef t_node name = lookupTypeDef name >>= \ty -> return (TypeDefRef name (Just ty) t_node)
 
 -- extract a struct\/union
@@ -293,7 +335,8 @@ tEnumTypeDecl handle_def (CEnum ident_opt enumerators_opt attrs node_info)
              return decl
 
 -- | translate and analyse an enumeration type
-tEnumType :: (MonadTrav m) => SUERef -> [(Ident, Maybe CExpr)] -> Attributes -> NodeInfo -> m EnumType
+tEnumType :: (MonadCError m, MonadSymtab m) =>
+             SUERef -> [(Ident, Maybe CExpr)] -> Attributes -> NodeInfo -> m EnumType
 tEnumType sue_ref enumerators attrs node = do
     mapM_ handleEnumeratorDef enumerators'
     return ty
@@ -307,11 +350,11 @@ tEnumType sue_ref enumerators attrs node = do
     nextEnrExpr (Left i) Nothing = (Left (succ i), intExpr i)
     nextEnrExpr (Right (e,offs)) Nothing = (Right (e, succ offs), offsExpr e offs)
     nextEnrExpr _ (Just e) = (Right (e,1), e)
-    intExpr i = CConst (CIntConst (cInteger i) internalNode)
-    offsExpr e offs = CBinary CAddOp e (intExpr offs) internalNode
+    intExpr i = CConst (CIntConst (cInteger i) undefNode)
+    offsExpr e offs = CBinary CAddOp e (intExpr offs) undefNode
 
 -- | Mapping from num type specs to C types (C99 6.7.2-2), ignoring the complex qualifier.
-tNumType :: (MonadTrav m) => NumTypeSpec -> m (Either (FloatType,Bool) IntType)
+tNumType :: (MonadCError m) => NumTypeSpec -> m (Either (FloatType,Bool) IntType)
 tNumType (NumTypeSpec basetype sgn sz iscomplex) =
     case (basetype,sgn,sz) of
         (BaseChar,_,NoSizeMod)      | Signed <- sgn   -> intType TySChar
@@ -372,7 +415,9 @@ data SizeMod     = NoSizeMod | ShortMod | LongMod | LongLongMod deriving (Eq,Ord
 data NumTypeSpec = NumTypeSpec { base :: NumBaseType, signSpec :: SignSpec, sizeMod :: SizeMod, isComplex :: Bool  }
 emptyNumTypeSpec :: NumTypeSpec
 emptyNumTypeSpec = NumTypeSpec { base = NoBaseType, signSpec = NoSignSpec, sizeMod = NoSizeMod, isComplex = False }
-data TypeSpecAnalysis = TSNone | TSVoid | TSBool | TSNum NumTypeSpec | TSNonBasic CTypeSpec
+data TypeSpecAnalysis =
+  TSNone | TSVoid | TSBool | TSNum NumTypeSpec |
+  TSTypeDef TypeDefRef | TSType Type | TSNonBasic CTypeSpec
 
 canonicalTypeSpec :: (MonadTrav m) => [CTypeSpec] -> m TypeSpecAnalysis
 canonicalTypeSpec = foldrM go TSNone where
@@ -382,6 +427,10 @@ canonicalTypeSpec = foldrM go TSNone where
     updLongMod NoSizeMod = Just LongMod
     updLongMod LongMod   = Just LongLongMod
     updLongMod _         = Nothing
+    getTypeSpecs :: MonadTrav m => Type -> m [CTypeSpec]
+    getTypeSpecs         =  return . getTS . partitionDeclSpecs . fst . exportType
+    getTS (_, _, _, ts, _) = ts
+    go :: (MonadTrav m) => CTypeSpec -> TypeSpecAnalysis -> m TypeSpecAnalysis
     go (CVoidType _)    TSNone = return$  TSVoid
     go (CBoolType _)    TSNone = return$  TSBool
     go (CCharType _)    tsa | (Just nts@(NumTypeSpec { base = NoBaseType })) <- getNTS tsa
@@ -403,11 +452,14 @@ canonicalTypeSpec = foldrM go TSNone where
                             = return$  TSNum$ nts { signSpec = Unsigned }
     go (CComplexType _) tsa | (Just nts@(NumTypeSpec { isComplex = False })) <- getNTS tsa
                             = return$  TSNum$ nts { isComplex = True }
+    go (CTypeDef i ni) TSNone = liftM TSTypeDef $ typeDefRef ni i
+    go (CTypeOfType d ni) TSNone = liftM TSType $ analyseTypeDecl d
+    go (CTypeOfExpr e _) TSNone = liftM TSType $ tExpr [] RValue e
     go otherType  TSNone    = return$  TSNonBasic otherType
     go ty _ts = astError (nodeInfo ty) "Invalid type specifier"
 
 -- compute storage given storage specifiers
-canonicalStorageSpec :: (MonadTrav m) =>[CStorageSpec] -> m StorageSpec
+canonicalStorageSpec :: (MonadCError m) =>[CStorageSpec] -> m StorageSpec
 canonicalStorageSpec storagespecs = liftM elideAuto $ foldrM updStorage NoStorageSpec storagespecs where
         updStorage (CAuto _) NoStorageSpec     = return$ AutoSpec
         updStorage (CRegister _) NoStorageSpec = return$ RegSpec
@@ -437,7 +489,7 @@ canonicalStorageSpec storagespecs = liftM elideAuto $ foldrM updStorage NoStorag
 -- > int f(int d, int c, char a, char* b)
 --
 -- TODO: This could be moved to syntax, as it operates on the AST only
-mergeOldStyle :: (MonadTrav m) => NodeInfo -> [CDecl] -> [CDerivedDeclr] -> m [CDerivedDeclr]
+mergeOldStyle :: (MonadCError m) => NodeInfo -> [CDecl] -> [CDerivedDeclr] -> m [CDerivedDeclr]
 mergeOldStyle _node [] declrs = return declrs
 mergeOldStyle node oldstyle_params (CFunDeclr params attrs fdnode : dds) =
     case params of
@@ -472,7 +524,7 @@ mergeOldStyle node _ _ = astError node "oldstyle parameter list, but not functio
 -- > [ [d| struct x { int z; } a, struct x *b |] ]
 --
 -- /TODO/: This could be moved to syntax, as it operates on the AST only
-splitCDecl :: (MonadTrav m) => CDecl -> m [CDecl]
+splitCDecl :: (MonadCError m) => CDecl -> m [CDecl]
 splitCDecl decl@(CDecl declspecs declrs node) =
     case declrs of
         []      -> internalErr "splitCDecl applied to empty declaration"
@@ -493,24 +545,25 @@ splitCDecl decl@(CDecl declspecs declrs node) =
 
 -- | translate @__attribute__@ annotations
 -- TODO: This is a unwrap and wrap stub
-tAttr :: (MonadTrav m) => CAttr -> m Attr
+tAttr :: (MonadCError m, MonadSymtab m) => CAttr -> m Attr
 tAttr (CAttr name cexpr node) = return$ Attr name cexpr node
 
 
 -- | construct a name for a variable
 -- TODO: more or less bogus
-mkVarName :: (MonadTrav m) => NodeInfo -> Maybe Ident -> Maybe AsmName -> m VarName
+mkVarName :: (MonadCError m, MonadSymtab m) =>
+             NodeInfo -> Maybe Ident -> Maybe AsmName -> m VarName
 mkVarName  node Nothing _ = return NoName
-mkVarName  node (Just n) asm = return $ VarName n Nothing
+mkVarName  node (Just n) asm = return $ VarName n asm
 
 -- helpers
-nameOfDecl :: (MonadTrav m) => CDecl -> m Ident
+nameOfDecl :: (MonadCError m) => CDecl -> m Ident
 nameOfDecl d = getOnlyDeclr d >>= \declr ->
     case declr of
         (CDeclr (Just name) _ _ _ _) -> return name
         (CDeclr Nothing _ _ _ node) -> internalErr "nameOfDecl: abstract declarator"
 emptyDeclr :: NodeInfo -> CDeclr
 emptyDeclr node = CDeclr Nothing [] Nothing [] node
-getOnlyDeclr :: (MonadTrav m) => CDecl -> m CDeclr
+getOnlyDeclr :: (MonadCError m) => CDecl -> m CDeclr
 getOnlyDeclr (CDecl _ [(Just declr,_,_)] _) = return declr
 getOnlyDeclr (CDecl _ _ node) = internalErr "getOnlyDeclr: declaration doesn't have a unique declarator"
