@@ -16,6 +16,12 @@
 -- Furthermore, the user may provide callbacks to handle declarations and definitions.
 -----------------------------------------------------------------------------
 module Language.C.Analysis.TravMonad (
+    -- * Name generation monad
+    MonadName(..),
+    -- * Symbol table monad
+    MonadSymtab(..),
+    -- * Specialized C error-handling monad
+    MonadCError(..),
     -- * AST traversal monad
     MonadTrav(..),
     -- * Handling declarations
@@ -37,54 +43,64 @@ module Language.C.Analysis.TravMonad (
     Trav,
     runTrav,runTrav_,
     TravState,initTravState,withExtDeclHandler,modifyUserState,userState,
+    getUserState,
+    TravOptions(..),modifyOptions,
     travErrors,
+    -- * Language options
+    CLanguage(..),
     -- * Helpers
     mapMaybeM,maybeM,mapSndM,concatMapM,
 )
 where
 import Language.C.Data
+import Language.C.Data.Ident
 import Language.C.Data.RList as RList
 import Language.C.Syntax
 
+import Language.C.Analysis.Builtins
 import Language.C.Analysis.SemError
 import Language.C.Analysis.SemRep
 import Language.C.Analysis.DefTable hiding (enterBlockScope,leaveBlockScope,
                                             enterFunctionScope,leaveFunctionScope)
 import qualified Language.C.Analysis.DefTable as ST
 
+import Data.IntMap (insert, lookup)
 import Data.Maybe
 import Control.Monad(liftM)
+import Prelude hiding (lookup)
 
--- | Traversal monad
-class (Monad m) => MonadTrav m where
+class (Monad m) => MonadName m where
+    -- | unique name generation
+    genName :: m Name
+
+class (Monad m) => MonadSymtab m where
+    -- symbol table handling
+    -- | return the definition table
+    getDefTable :: m DefTable
+    -- | perform an action modifying the definition table
+    withDefTable :: (DefTable -> (a, DefTable)) -> m a
+
+class (Monad m) => MonadCError m where
     -- error handling facilities
 
     -- | throw an 'Error'
     throwTravError :: Error e => e -> m a
     -- | catch an 'Error' (we could implement dynamically-typed catch here)
     catchTravError :: m a -> (CError -> m a) -> m a
-    -- | remember that an 'Error' occured (without throwing it)
+    -- | remember that an 'Error' occurred (without throwing it)
     recordError    :: Error e => e -> m ()
     -- | return the list of recorded errors
     getErrors      :: m [CError]
 
-    -- symbol table handling
-
-    -- | return the definition table
-    getDefTable :: m DefTable
-    -- | perform an action modifying the definition table
-    withDefTable :: (DefTable -> (a, DefTable)) -> m a
-
-    -- | unique name generation
-    genName :: m Name
-
+-- | Traversal monad
+class (MonadName m, MonadSymtab m, MonadCError m) => MonadTrav m where
     -- | handling declarations and definitions
     handleDecl :: DeclEvent -> m ()
 
 -- * handling declarations
 
 -- check wheter a redefinition is ok
-checkRedef :: (MonadTrav m, CNode t, CNode t1) => String -> t -> (DeclarationStatus t1) -> m ()
+checkRedef :: (MonadCError m, CNode t, CNode t1) => String -> t -> (DeclarationStatus t1) -> m ()
 checkRedef subject new_decl redecl_status =
     case redecl_status of
         NewDecl -> return ()
@@ -97,9 +113,9 @@ checkRedef subject new_decl redecl_status =
             -- redefinition LevelWarn subject ShadowedDef (nodeInfo new_decl) (nodeInfo old_def)
         KeepDef _old_def      -> return ()
 
--- | forward declaration of a tag. Only neccessary for name analysis, but otherwise no semantic
+-- | forward declaration of a tag. Only necessary for name analysis, but otherwise no semantic
 -- consequences.
-handleTagDecl :: (MonadTrav m) => TagFwdDecl -> m ()
+handleTagDecl :: (MonadCError m, MonadSymtab m) => TagFwdDecl -> m ()
 handleTagDecl decl = do
     redecl <- withDefTable $ declareTag (sueRef decl) decl
     checkRedef (show $ sueRef decl) decl redecl
@@ -114,7 +130,7 @@ handleTagDef def = do
     checkRedef (show $ sueRef def) def redecl
     handleDecl (TagEvent def)
 
-handleEnumeratorDef :: (MonadTrav m) => Enumerator ->  m ()
+handleEnumeratorDef :: (MonadCError m, MonadSymtab m) => Enumerator ->  m ()
 handleEnumeratorDef enumerator = do
     let ident = declIdent enumerator
     redecl <- withDefTable $ defineScopedIdent ident (EnumeratorDef enumerator)
@@ -131,12 +147,13 @@ handleTypeDef typeDef@(TypeDef ident _ _ _) = do
 handleAsmBlock :: (MonadTrav m) => AsmBlock -> m ()
 handleAsmBlock asm = handleDecl (AsmEvent asm)
 
-redefErr :: (MonadTrav m, CNode old, CNode new) => Ident -> ErrorLevel -> new -> old -> RedefKind -> m ()
+redefErr :: (MonadCError m, CNode old, CNode new) =>
+            Ident -> ErrorLevel -> new -> old -> RedefKind -> m ()
 redefErr name lvl new old kind =
   throwTravError $ redefinition lvl (show name) kind (nodeInfo new) (nodeInfo old)
 
 -- TODO: unused
-checkIdentTyRedef :: (MonadTrav m) => IdentEntry -> (DeclarationStatus IdentEntry) -> m ()
+checkIdentTyRedef :: (MonadCError m) => IdentEntry -> (DeclarationStatus IdentEntry) -> m ()
 checkIdentTyRedef (Right decl) status = checkVarRedef decl status
 checkIdentTyRedef (Left tydef) (KindMismatch old_def) =
   redefErr (identOfTypeDef tydef) LevelError tydef old_def DiffKindRedecl
@@ -144,49 +161,67 @@ checkIdentTyRedef (Left tydef) (Redeclared old_def) =
   redefErr (identOfTypeDef tydef) LevelError tydef old_def DuplicateDef
 checkIdentTyRedef (Left _tydef) _ = return ()
 
-checkVarRedef :: (MonadTrav m) => IdentDecl -> (DeclarationStatus IdentEntry) -> m ()
+-- Check whether it is ok to declare a variable already in scope
+checkVarRedef :: (MonadCError m) => IdentDecl -> (DeclarationStatus IdentEntry) -> m ()
 checkVarRedef def redecl =
     case redecl of
         -- always an error
         KindMismatch old_def -> redefVarErr old_def DiffKindRedecl
-        -- types have to match
-        KeepDef (Right old_def) -> throwOnLeft $ checkCompatibleTypes new_ty (declType old_def)
-        -- redeclaration: old entry has to be a declaration or tentative, type have to match
-        Redeclared (Right old_def) | isTentativeG old_def ->
-                                      throwOnLeft $ checkCompatibleTypes new_ty (declType old_def)
-                                   | otherwise -> redefVarErr old_def DuplicateDef
+        -- Declaration referencing definition:
+        --   * new entry has to be a declaration
+        --   * old entry and new entry have to have linkage and agree on linkage
+        --   * types have to match
+        KeepDef (Right old_def) | not (agreeOnLinkage def old_def) -> linkageErr def old_def
+                                | otherwise -> throwOnLeft $ checkCompatibleTypes new_ty (declType old_def)
+        -- redefinition:
+        --   * old entry has to be a declaration or tentative definition
+        --   * old entry and new entry have to have linkage and agree on linkage
+        --   * types have to match
+        Redeclared (Right old_def) | not (agreeOnLinkage def old_def) -> linkageErr def old_def
+                                   | not(canBeOverwritten old_def) -> redefVarErr old_def DuplicateDef
+                                   | otherwise -> throwOnLeft $ checkCompatibleTypes new_ty (declType old_def)
         -- NewDecl/Shadowed is ok
         _ -> return ()
     where
     redefVarErr old_def kind = redefErr (declIdent def) LevelError def old_def kind
+    linkageErr def old_def =
+        case (declLinkage def, declLinkage old_def) of
+            (NoLinkage, _) -> redefErr (declIdent def) LevelError  def old_def NoLinkageOld
+            otherwise      -> redefErr (declIdent def) LevelError  def old_def DisagreeLinkage
+
     new_ty = declType def
-    isTentativeG (Declaration _) = True
-    isTentativeG (ObjectDef od)  = isTentative od
-    isTentativeG _               = False
+    canBeOverwritten (Declaration _) = True
+    canBeOverwritten (ObjectDef od)  = isTentative od
+    canBeOverwritten _               = False
+    agreeOnLinkage def old_def
+        | declStorage old_def == FunLinkage InternalLinkage = True
+        | not (hasLinkage $ declStorage def) || not (hasLinkage $ declStorage old_def) = False
+        | (declLinkage def) /= (declLinkage old_def) = False
+        | otherwise = True
 
 -- | handle variable declarations (external object declarations and function prototypes)
 -- variable declarations are either function prototypes, or external declarations, and not very
 -- interesting on their own. we only put them in the symbol table and call the handle.
 -- declarations never override definitions
-handleVarDecl :: (MonadTrav m) => Decl -> m ()
-handleVarDecl decl = do
+handleVarDecl :: (MonadTrav m) => Bool -> Decl -> m ()
+handleVarDecl is_local decl = do
     def <- enterDecl decl (const False)
-    handleDecl (DeclEvent def)
+    handleDecl ((if is_local then LocalEvent else DeclEvent) def)
 
 -- | handle parameter declaration. The interesting part is that parameters can be abstract
 -- (if they are part of a type). If they have a name, we enter the name (usually in function prototype or function scope),
--- checking if there are duplicate definitions. 
+-- checking if there are duplicate definitions.
 -- FIXME: I think it would be more transparent to handle parameter declarations in a special way
 handleParamDecl :: (MonadTrav m) => ParamDecl -> m ()
-handleParamDecl (AbstractParamDecl _ _) = return ()
-handleParamDecl (ParamDecl vardecl node) = do
+handleParamDecl pd@(AbstractParamDecl _ _) = handleDecl (ParamEvent pd)
+handleParamDecl pd@(ParamDecl vardecl node) = do
     let def = ObjectDef (ObjDef vardecl Nothing node)
     redecl <- withDefTable $ defineScopedIdent (declIdent def) def
     checkVarRedef def redecl
-    return ()
+    handleDecl (ParamEvent pd)
 
 -- shared impl
-enterDecl :: (MonadTrav m) => Decl -> (IdentDecl -> Bool) -> m IdentDecl
+enterDecl :: (MonadCError m, MonadSymtab m) => Decl -> (IdentDecl -> Bool) -> m IdentDecl
 enterDecl decl cond = do
     let def = Declaration decl
     redecl <- withDefTable $
@@ -211,20 +246,20 @@ checkCompatibleTypes :: Type -> Type -> Either TypeMismatch ()
 checkCompatibleTypes _ _ = Right ()
 
 -- | handle object defintions (maybe tentative)
-handleObjectDef :: (MonadTrav m) => Ident -> ObjDef -> m ()
-handleObjectDef ident obj_def = do
+handleObjectDef :: (MonadTrav m) => Bool -> Ident -> ObjDef -> m ()
+handleObjectDef local ident obj_def = do
     let def = ObjectDef obj_def
     redecl <- withDefTable $
-              defineScopedIdentWhen (\o -> shouldOverride def o) ident def
+              defineScopedIdentWhen (\old -> shouldOverride def old) ident def
     checkVarRedef def redecl
-    handleDecl (DeclEvent def)
+    handleDecl ((if local then LocalEvent else DeclEvent) def)
     where
     isTentativeDef (ObjectDef object_def) = isTentative object_def
     isTentativeDef _ = False
-    shouldOverride def o | isDeclaration o = True
-                         | not (isTentativeDef def) = True
-                         | isTentativeDef o = True
-                         | otherwise = False
+    shouldOverride def old | isDeclaration old = True
+                           | not (isTentativeDef def) = True
+                           | isTentativeDef old = True
+                           | otherwise = False
 
 -- * scope manipulation
 --
@@ -235,25 +270,25 @@ handleObjectDef ident obj_def = do
 --  * function scope: labels are visible within the entire function, and declared implicitely
 --
 --  * block scope
-updDefTable :: (MonadTrav m) => (DefTable -> DefTable) -> m ()
+updDefTable :: (MonadSymtab m) => (DefTable -> DefTable) -> m ()
 updDefTable f = withDefTable (\st -> ((),f st))
 
-enterPrototypeScope :: (MonadTrav m) => m ()
+enterPrototypeScope :: (MonadSymtab m) => m ()
 enterPrototypeScope = updDefTable (ST.enterBlockScope)
 
-leavePrototypeScope :: (MonadTrav m) => m ()
+leavePrototypeScope :: (MonadSymtab m) => m ()
 leavePrototypeScope = updDefTable (ST.leaveBlockScope)
 
-enterFunctionScope :: (MonadTrav m) => m ()
+enterFunctionScope :: (MonadSymtab m) => m ()
 enterFunctionScope = updDefTable (ST.enterFunctionScope)
 
-leaveFunctionScope :: (MonadTrav m) => m ()
+leaveFunctionScope :: (MonadSymtab m) => m ()
 leaveFunctionScope = updDefTable (ST.leaveFunctionScope)
 
-enterBlockScope :: (MonadTrav m) => m ()
+enterBlockScope :: (MonadSymtab m) => m ()
 enterBlockScope = updDefTable (ST.enterBlockScope)
 
-leaveBlockScope :: (MonadTrav m) => m ()
+leaveBlockScope :: (MonadSymtab m) => m ()
 leaveBlockScope = updDefTable (ST.leaveBlockScope)
 
 -- * Lookup
@@ -262,26 +297,41 @@ leaveBlockScope = updDefTable (ST.leaveBlockScope)
 -- the 'wrong kind of object' is an internal error here,
 -- because the parser should distinguish typeDefs and other
 -- objects
-lookupTypeDef :: (MonadTrav m) => Ident -> m Type
+lookupTypeDef :: (MonadCError m, MonadSymtab m) => Ident -> m Type
 lookupTypeDef ident =
     getDefTable >>= \symt ->
     case lookupIdent ident symt of
-        Nothing                             -> astError (nodeInfo ident) "unbound typeDef"
-        Just (Left (TypeDef _ident ty _ _)) -> return ty
+        Nothing                             ->
+          astError (nodeInfo ident) $ "unbound typeDef: " ++ identToString ident
+        Just (Left (TypeDef def_ident ty _ _)) -> addRef ident def_ident >> return ty
         Just (Right d)                      -> astError (nodeInfo ident) (wrongKindErrMsg d)
     where
-    wrongKindErrMsg d = "wrong kind of object: excepcted typeDef but found: "++(objKindDescr d)
+    wrongKindErrMsg d = "wrong kind of object: expected typedef but found "++ (objKindDescr d)
+                        ++ " (for identifier `" ++ identToString ident ++ "')"
 
 
 -- | lookup an object, function or enumerator
-lookupObject :: (MonadTrav m) => Ident -> m (Maybe IdentDecl)
+lookupObject :: (MonadCError m, MonadSymtab m) => Ident -> m (Maybe IdentDecl)
 lookupObject ident = do
     old_decl <- liftM (lookupIdent ident) getDefTable
     mapMaybeM old_decl $ \obj ->
         case obj of
-        Right objdef -> return objdef
+        Right objdef -> addRef ident objdef >> return objdef
         Left _tydef  -> astError (nodeInfo ident) (mismatchErr "lookupObject" "an object" "a typeDef")
 
+-- | add link between use and definition (private)
+addRef :: (MonadCError m, MonadSymtab m, CNode u, CNode d) => u -> d -> m ()
+addRef use def =
+  case (nodeInfo use, nodeInfo def) of
+    (NodeInfo _ _ useName, NodeInfo _ _ defName) ->
+      withDefTable
+      (\dt ->
+         ((),
+          dt { refTable = insert (nameId useName) defName (refTable dt) }
+         )
+      )
+    (_, _) -> return () -- Don't have Names for both, so can't record.
+
 
 mismatchErr :: String -> String -> String -> String
 mismatchErr ctx expect found = ctx ++ ": Expected " ++ expect ++ ", but found: " ++ found
@@ -293,30 +343,30 @@ mismatchErr ctx expect found = ctx ++ ": Expected " ++ expect ++ ", but found: "
 -- This currently depends on the fact the structs are tagged with unique names.
 -- We could use the name generation of TravMonad as well, which might be the better
 -- choice when dealing with autogenerated code.
-createSUERef :: (MonadTrav m) => NodeInfo -> Maybe Ident -> m SUERef
+createSUERef :: (MonadCError m, MonadSymtab m) => NodeInfo -> Maybe Ident -> m SUERef
 createSUERef _node_info (Just ident) = return$ NamedRef ident
 createSUERef node_info Nothing | (Just name) <- nameOfNode node_info = return $ AnonymousRef name
                                | otherwise = astError node_info "struct/union/enum definition without unique name"
 
 -- * error handling facilities
 
-handleTravError :: (MonadTrav m) => m a -> m (Maybe a)
+handleTravError :: (MonadCError m) => m a -> m (Maybe a)
 handleTravError a = liftM Just a `catchTravError` (\e -> recordError e >> return Nothing)
 
--- | check wheter non-recoverable errors occured
+-- | check wheter non-recoverable errors occurred
 hadHardErrors :: [CError] -> Bool
 hadHardErrors = (not . null . filter isHardError)
 
 -- | raise an error caused by a malformed AST
-astError :: (MonadTrav m) => NodeInfo -> String -> m a
+astError :: (MonadCError m) => NodeInfo -> String -> m a
 astError node msg = throwTravError $ invalidAST node msg
 
 -- | raise an error based on an Either argument
-throwOnLeft :: (MonadTrav m, Error e) => Either e a -> m a
+throwOnLeft :: (MonadCError m, Error e) => Either e a -> m a
 throwOnLeft (Left err) = throwTravError err
 throwOnLeft (Right v)  = return v
 
-warn :: (Error e, MonadTrav m) => e -> m ()
+warn :: (Error e, MonadCError m) => e -> m ()
 warn err = recordError (changeErrorLevel err LevelWarn)
 
 -- * The Trav datatype
@@ -332,6 +382,7 @@ get      = Trav (\s -> Right (s,s))
 put :: TravState s -> Trav s ()
 put s    = Trav (\_ -> Right ((),s))
 
+
 runTrav :: forall s a. s -> Trav s a -> Either [CError] (a, TravState s)
 runTrav state traversal =
     case unTrav action (initTravState state) of
@@ -339,12 +390,8 @@ runTrav state traversal =
         Right (v, ts) | hadHardErrors (travErrors ts) -> Left (travErrors ts)
                       | otherwise                     -> Right (v,ts)
     where
-    action = do withDefTable $ defineTypeDef (identOfTypeDef va_list) va_list
+    action = do withDefTable (const ((), builtins))
                 traversal
-    va_list = (TypeDef (internalIdent "__builtin_va_list")
-                       (DirectType (TyBuiltin TyVaList) noTypeQuals)
-                       []
-                       (internalNode))
 
 runTrav_ :: Trav () a -> Either [CError] (a,[CError])
 runTrav_ t = fmap fst . runTrav () $
@@ -363,14 +410,11 @@ instance Monad (Trav s) where
                               Right (x,s1) -> unTrav (k x) s1
                               Left e       -> Left e)
 
-instance MonadTrav (Trav s) where
-    -- error handling facilities
-    throwTravError e = Trav (\_ -> Left (toError e))
-    catchTravError a handler = Trav (\s -> case unTrav a s of
-                                             Left e  -> unTrav (handler e) s
-                                             Right r -> Right r)
-    recordError e = modify $ \st -> st { rerrors = (rerrors st) `snoc` toError e }
-    getErrors = gets (RList.reverse . rerrors)
+instance MonadName (Trav s) where
+    -- unique name generation
+    genName = generateName
+
+instance MonadSymtab (Trav s) where
     -- symbol table handling
     getDefTable = gets symbolTable
     withDefTable f = do
@@ -378,18 +422,36 @@ instance MonadTrav (Trav s) where
         let (r,symt') = f (symbolTable ts)
         put $ ts { symbolTable = symt' }
         return r
-    -- unique name generation
-    genName = generateName
+
+instance MonadCError (Trav s) where
+    -- error handling facilities
+    throwTravError e = Trav (\_ -> Left (toError e))
+    catchTravError a handler = Trav (\s -> case unTrav a s of
+                                             Left e  -> unTrav (handler e) s
+                                             Right r -> Right r)
+    recordError e = modify $ \st -> st { rerrors = (rerrors st) `snoc` toError e }
+    getErrors = gets (RList.reverse . rerrors)
+
+instance MonadTrav (Trav s) where
     -- handling declarations and definitions
     handleDecl d = ($ d) =<< gets doHandleExtDecl
 
+-- | The variety of the C language to accept. Note: this is not yet enforced.
+data CLanguage = C89 | C99 | GNU89 | GNU99
+
+data TravOptions =
+    TravOptions {
+        language :: CLanguage
+    }
+
 data TravState s =
     TravState {
         symbolTable :: DefTable,
         rerrors :: RList CError,
         nameGenerator :: [Name],
         doHandleExtDecl :: (DeclEvent -> Trav s ()),
-        userState :: s
+        userState :: s,
+        options :: TravOptions
       }
 
 travErrors :: TravState s -> [CError]
@@ -402,13 +464,20 @@ initTravState userst =
         rerrors = RList.empty,
         nameGenerator = newNameSupply,
         doHandleExtDecl = const (return ()),
-        userState = userst
+        userState = userst,
+        options = TravOptions { language = C99 }
       }
 
 -- * Trav specific operations
 modifyUserState :: (s -> s) -> Trav s ()
 modifyUserState f = modify $ \ts -> ts { userState = f (userState ts) }
 
+getUserState :: Trav s s
+getUserState = userState `liftM` get
+
+modifyOptions :: (TravOptions -> TravOptions) -> Trav s ()
+modifyOptions f = modify $ \ts -> ts { options = f (options ts) }
+
 generateName :: Trav s Name
 generateName =
     get >>= \ts ->
