@@ -10,7 +10,7 @@
 -- Portability :  ghc
 --
 -- This module provides the notion of identifiers in C, speed up using hashing.
--- Identifiers are associated optionally associated with a 'NodeInfo', i.e. with 
+-- Identifiers are associated optionally associated with a 'NodeInfo', i.e. with
 -- a unique 'Name' and a source location ('Position'). The ordering relation on
 -- identifiers is based on the hash and does not follow the lexical order.
 -----------------------------------------------------------------------------
@@ -19,7 +19,7 @@ module Language.C.Data.Ident (
     SUERef(..), isAnonymousRef,
     mkIdent, builtinIdent, internalIdent, internalIdentAt, isInternalIdent, identToString, dumpIdent)
 where
-  
+
 -- TODO (comment from manuel):
 --  * Hashing is not 8bit clean.
 
@@ -46,8 +46,8 @@ isAnonymousRef _ = False
 
 -- | C identifiers
 data Ident = Ident String       -- lexeme
-                   {-# UNPACK #-}   !Int         -- hash to speed up equality check
-                   NodeInfo        -- attributes of this ident. incl. position
+                   {-# UNPACK #-}   !Int     -- hash to speed up equality check
+                   NodeInfo                   -- attributes of this ident. incl. position
              deriving (Data,Typeable)
 
 -- the definition of the equality allows identifiers to be equal that are
@@ -64,7 +64,7 @@ instance Ord Ident where
 
 -- for displaying identifiers
 instance Show Ident where
-  showsPrec _ ide = showString ("`" ++ identToString ide ++ "'")
+  showsPrec _ ide = showString ("\"" ++ identToString ide ++ "\"")
 
 -- identifiers are attributed
 instance CNode Ident where
@@ -108,7 +108,7 @@ bits28 = 2^(28::Int)
 --
 -- * for reasons of simplicity the complete lexeme is hashed.
 mkIdent            :: Position -> String -> Name -> Ident
-mkIdent pos s name  = Ident s (quad s) (mkNodeInfo pos name)
+mkIdent pos s name  = Ident s (quad s) (mkNodeInfo' pos (pos,length s) name)
 
 -- | returns an /internal/ identifier (has internal position and no unique name)
 internalIdent   :: String -> Ident
@@ -116,7 +116,7 @@ internalIdent s  = Ident s (quad s) (mkNodeInfoOnlyPos internalPos)
 
 -- | return an /internal/ identifier with position info
 internalIdentAt :: Position -> String -> Ident
-internalIdentAt pos s = Ident s (quad s) (mkNodeInfoOnlyPos pos)
+internalIdentAt pos s = Ident s (quad s) (mkNodeInfoPosLen pos (pos, length s))
 
 -- | returns a /builtin/ identifier (has builtin position and no unique name)
 builtinIdent   :: String -> Ident
