@@ -1,4 +1,4 @@
-{-# OPTIONS  #-}
+{-# LANGUAGE TypeSynonymInstances, FlexibleInstances  #-}
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Language.C.Pretty
@@ -73,7 +73,7 @@ instance Pretty CTranslUnit where
 
 -- | Pretty print the given tranlation unit, but replace declarations from header files with @#include@ directives.
 --
--- The resulting file may not compile (because of missing @#define@ directives and similar things), but is very usefull
+-- The resulting file may not compile (because of missing @#define@ directives and similar things), but is very useful
 -- for testing, as otherwise the pretty printed file will be cluttered with declarations from system headers.
 prettyUsingInclude :: CTranslUnit -> Doc
 prettyUsingInclude (CTranslUnit edecls _) =
@@ -91,13 +91,13 @@ prettyUsingInclude (CTranslUnit edecls _) =
     includeHeader hFile = text "#include" <+> doubleQuotes (text hFile)
     isHeaderFile = (".h" `isSuffixOf`)
     includeWarning hs | Set.null hs = empty
-                      | otherwise = text "/* Warning: The #include directives in this file aren't neccessarily correct. */"
+                      | otherwise = text "/* Warning: The #include directives in this file aren't necessarily correct. */"
 
 -- TODO: Check need of __extension__
 instance Pretty CExtDecl where
     pretty (CDeclExt decl) = pretty decl <> semi
     pretty (CFDefExt fund) = pretty fund
-    pretty (CAsmExt  asmStmt) = text "asm" <> parens (pretty asmStmt) <> semi
+    pretty (CAsmExt  asmStmt _) = text "asm" <> parens (pretty asmStmt) <> semi
 
 -- TODO: Check that old-style and new-style aren't mixed
 instance Pretty CFunDef where
@@ -119,18 +119,19 @@ instance Pretty CStat where
     pretty (CExpr expr _) = ii $ maybeP pretty expr <> semi
     pretty c@(CCompound _ _ _) = prettyPrec 0 c
     pretty (CIf expr stat estat _) =
-        ii $  text "if" <+> text "(" <> pretty expr <> text ")"
-                $+$ prettyThen stat
+        ii $  text "if" <+> parens (pretty expr)
+                $+$ prettyBody stat
                 $$  maybeP prettyElse estat
       where
-        prettyThen stat@(CIf _ _ _ _) = text "{" $+$ prettyPrec (-1) stat $$ text "}"
-        prettyThen stat = prettyPrec (-1) stat
+        prettyBody c@(CCompound _ _ _) = prettyPrec (-1) c
+        prettyBody nonCompound         = prettyPrec (-1) (CCompound [] [CBlockStmt nonCompound] undefined)
         prettyElse (CIf else_if_expr else_if_stat else_stat _) =
-          text "else if" <+> text "(" <> pretty else_if_expr <> text ")"
-            $+$ prettyPrec (-1) else_if_stat
-          $$ maybeP prettyElse else_stat
+          text "else if" <+> parens (pretty else_if_expr)
+            $+$ prettyBody else_if_stat
+            $$  maybeP prettyElse else_stat
         prettyElse else_stmt =
-          text "else" $+$ prettyPrec (-1) else_stmt
+          text "else"
+            $+$ prettyBody else_stmt
 
     pretty (CSwitch expr stat _) =
         ii $ text "switch" <+> text "(" <> pretty expr <> text ")"
@@ -384,7 +385,7 @@ instance Pretty CExpr where
     prettyPrec p (CUnary CPostDecOp expr _) =
         parenPrec p 26 $ prettyPrec 26 expr <> text "--"
     prettyPrec p (CUnary op expr@(CUnary _ _ _) _) =
-        --                             parens aren't neccessary, but look nicer imho
+        --                             parens aren't necessary, but look nicer imho
         parenPrec p 25 $ pretty op <+> parens (prettyPrec 25 expr)
     prettyPrec p (CUnary op expr _) =
         parenPrec p 25 $ pretty op <> prettyPrec 25 expr
@@ -439,13 +440,52 @@ instance Pretty CBuiltin where
         (parens $ pretty ty1 <> comma <+> pretty ty2)
 
 instance Pretty CAssignOp where
-    pretty = text . show
+  pretty op = text $ case op of
+    CAssignOp -> "="
+    CMulAssOp -> "*="
+    CDivAssOp -> "/="
+    CRmdAssOp -> "%="
+    CAddAssOp -> "+="
+    CSubAssOp -> "-="
+    CShlAssOp -> "<<="
+    CShrAssOp -> ">>="
+    CAndAssOp -> "&="
+    CXorAssOp -> "^="
+    COrAssOp  -> "|="
 
 instance Pretty CBinaryOp where
-    pretty = text . show
+  pretty op = text $ case op of
+    CMulOp -> "*"
+    CDivOp -> "/"
+    CRmdOp -> "%"
+    CAddOp -> "+"
+    CSubOp -> "-"
+    CShlOp -> "<<"
+    CShrOp -> ">>"
+    CLeOp  -> "<"
+    CGrOp  -> ">"
+    CLeqOp -> "<="
+    CGeqOp -> ">="
+    CEqOp  -> "=="
+    CNeqOp -> "!="
+    CAndOp -> "&"
+    CXorOp -> "^"
+    COrOp  -> "|"
+    CLndOp -> "&&"
+    CLorOp -> "||"
 
 instance Pretty CUnaryOp where
-    pretty = text . show
+  pretty op = text $ case op of
+    CPreIncOp  -> "++"
+    CPreDecOp  -> "--"
+    CPostIncOp -> "++"
+    CPostDecOp -> "--"
+    CAdrOp     -> "&"
+    CIndOp     -> "*"
+    CPlusOp    -> "+"
+    CMinOp     -> "-"
+    CCompOp    -> "~"
+    CNegOp     -> "!"
 
 instance Pretty CConst where
     pretty (CIntConst   int_const _) = text (show int_const)
